@@ -39,11 +39,18 @@ import {
   Deriv,
   Exp,
   Floor,
+  HistogramAvg,
+  HistogramCount,
+  HistogramFraction,
   HistogramQuantile,
+  HistogramStdDev,
+  HistogramStdVar,
+  HistogramSum,
   HoltWinters,
   Hour,
   Idelta,
   Increase,
+  Info,
   Irate,
   LabelJoin,
   LabelReplace,
@@ -51,6 +58,7 @@ import {
   Ln,
   Log10,
   Log2,
+  MadOverTime,
   MaxOverTime,
   MinOverTime,
   Minute,
@@ -69,6 +77,8 @@ import {
   Sinh,
   Sort,
   SortDesc,
+  SortByLabel,
+  SortByLabelDesc,
   Sqrt,
   StddevOverTime,
   StdvarOverTime,
@@ -261,12 +271,48 @@ const promqlFunctions: { [key: number]: PromQLFunction } = {
     variadic: 0,
     returnType: ValueType.vector,
   },
+  [HistogramAvg]: {
+    name: 'histogram_avg',
+    argTypes: [ValueType.vector],
+    variadic: 0,
+    returnType: ValueType.vector,
+  },
+  [HistogramCount]: {
+    name: 'histogram_count',
+    argTypes: [ValueType.vector],
+    variadic: 0,
+    returnType: ValueType.vector,
+  },
+  [HistogramFraction]: {
+    name: 'histogram_fraction',
+    argTypes: [ValueType.scalar, ValueType.scalar, ValueType.vector],
+    variadic: 0,
+    returnType: ValueType.vector,
+  },
   [HistogramQuantile]: {
     name: 'histogram_quantile',
     argTypes: [ValueType.scalar, ValueType.vector],
     variadic: 0,
     returnType: ValueType.vector,
   },
+  [HistogramStdDev]: {
+    name: 'histogram_stddev',
+    argTypes: [ValueType.vector],
+    variadic: 0,
+    returnType: ValueType.vector,
+  },
+  [HistogramStdVar]: {
+    name: 'histogram_stdvar',
+    argTypes: [ValueType.vector],
+    variadic: 0,
+    returnType: ValueType.vector,
+  },
+  [HistogramSum]: {
+    name: 'histogram_sum',
+    argTypes: [ValueType.vector],
+    variadic: 0,
+    returnType: ValueType.vector,
+  },
   [HoltWinters]: {
     name: 'holt_winters',
     argTypes: [ValueType.matrix, ValueType.scalar, ValueType.scalar],
@@ -291,6 +337,12 @@ const promqlFunctions: { [key: number]: PromQLFunction } = {
     variadic: 0,
     returnType: ValueType.vector,
   },
+  [Info]: {
+    name: 'info',
+    argTypes: [ValueType.vector, ValueType.vector],
+    variadic: 1,
+    returnType: ValueType.vector,
+  },
   [Irate]: {
     name: 'irate',
     argTypes: [ValueType.matrix],
@@ -333,6 +385,12 @@ const promqlFunctions: { [key: number]: PromQLFunction } = {
     variadic: 0,
     returnType: ValueType.vector,
   },
+  [MadOverTime]: {
+    name: 'mad_over_time',
+    argTypes: [ValueType.matrix],
+    variadic: 0,
+    returnType: ValueType.vector,
+  },
   [MaxOverTime]: {
     name: 'max_over_time',
     argTypes: [ValueType.matrix],
@@ -441,6 +499,18 @@ const promqlFunctions: { [key: number]: PromQLFunction } = {
     variadic: 0,
     returnType: ValueType.vector,
   },
+  [SortByLabel]: {
+    name: 'sort_by_label',
+    argTypes: [ValueType.vector, ValueType.string],
+    variadic: -1,
+    returnType: ValueType.vector,
+  },
+  [SortByLabelDesc]: {
+    name: 'sort_by_label_desc',
+    argTypes: [ValueType.vector, ValueType.string],
+    variadic: -1,
+    returnType: ValueType.vector,
+  },
   [Sqrt]: {
     name: 'sqrt',
     argTypes: [ValueType.vector],
